/**
 * A hack to ape nominal typing.
 * @hidden
 */
declare type Nominal<Ty, Discriminant> = Ty & {
    __discriminant: Discriminant;
};
/**
 * Used to ensure that nothing else can invoke these classes' constructors directly; this mints a new, locally-scoped Symbol that cannot be reproduced outside of this file.
 * @hidden
 */
declare type sentinel = Nominal<symbol, 'INTERNAL'>;
declare type checkpoint_state = 'InputNeeded' | 'Shifting' | 'AboutToReduce' | 'HandlingError' | 'Accepted' | 'Rejected';
/** @hidden */
declare type $exn = Nominal<Array<any>, 'exn'>;
/** @hidden */
declare type $buffer = Nominal<object, 'Lexer.buffer'>;
/** @hidden */
declare type $token = Nominal<object, 'Tokens.token'>;
/** @hidden */
declare type $token_located = Nominal<object, 'Tokens.token located'>;
/** @hidden */
declare type $position = Nominal<object, 'Lexing.position'>;
/** @hidden */
declare type $ASTt = Nominal<object, 'AST.t'>;
/** @hidden */
declare type $Expressiont = Nominal<object, 'Expression.t'>;
/** @hidden */
declare type $checkpoint = Nominal<object, 'Incremental.checkpoint'>;
/** @hidden */
declare type $element = Nominal<object, 'MenhirInterpreter.element'>;
declare type ParseOptions = {
    throwException?: boolean;
};
export declare class OCamlError extends Error {
    /** @hidden */
    $exn: $exn;
    constructor($exn: $exn, message?: string);
}
export declare class LexError extends OCamlError {
    pos: Position;
    constructor($exn: $exn);
}
export declare class ParseError extends OCamlError {
    token: LocatedToken;
    constructor($exn: $exn);
}
/**
 * As opposed to [[Sub]]; represents simple, directly-embedded `string`s in the AST. These are
 * the values that are immediately available, and do not need further evaluation.
 *
 * (Functions that return this are usually also available as `eval*` versions that help you
 * interpret subexpressions recursively; see [[Expression]] for more information.)
 */
interface Literal {
    type: 'literal';
    value: string;
}
/**
 * As opposed to [[Literal]]; represents an sub-[[Expression]] embedded in the AST.
 *
 * (Functions that return this are usually also available as `eval*` versions that help you
 * interpret subexpressions recursively; see [[Expression]] for more information.)
 */
interface Sub<T> {
    type: 'subexpression';
    expr: T;
}
/**
 * The function-signature required of callbacks, passed to `eval*` functions and friends, that can
 * reduce an unevaluated sub-[[Expression]] into a resultant, simple `string`-ish return value.
 *
 * See [[Expression]] for more information.
 */
declare type evaluator = (expr: ExpressionEval) => string;
declare function script(lexbuf: LexBuffer, options?: ParseOptions): Script;
declare function scriptOfString(str: string, options?: ParseOptions): Script;
declare function startScript(lexbuf: LexBuffer): Checkpoint<"script">;
declare function startScriptWithString(str: string): Checkpoint<"script">;
declare function expression(lexbuf: LexBuffer, options?: ParseOptions): Expression;
declare function expressionOfString(str: string, options?: ParseOptions): Expression;
declare function startExpression(lexbuf: LexBuffer): Checkpoint<"expression">;
declare function startExpressionWithString(str: string): Checkpoint<"expression">;
export declare const Parser: {
    script: typeof script;
    scriptOfString: typeof scriptOfString;
    startScript: typeof startScript;
    startScriptWithString: typeof startScriptWithString;
    expression: typeof expression;
    expressionOfString: typeof expressionOfString;
    startExpression: typeof startExpression;
    startExpressionWithString: typeof startExpressionWithString;
};
export declare class Script {
    /** @hidden */
    private $scpt;
    /** @hidden */
    private $expressions;
    constructor(isInternal: sentinel, $scpt: $ASTt);
    readonly expressions: Expression[];
}
declare class ExpressionCommon {
    /** @hidden */
    protected $expr: $Expressiont;
    constructor(isInternal: sentinel, $expr: $Expressiont);
    readonly count: number;
    readonly flags: string[];
    hasFlag(flag: string): boolean;
}
/**
 * The core type of my AST, this data-structure represents a single (sub-)command in an excmd
 * script; and is the product of entry-points like [[Parser.expressionOfString]].
 *
 * ### Construction
 * This JavaScript class, like most in my interface, cannot be constructed directly (i.e. `new
 * Expression()` is going to throw an error); instead, you must invoke the parser via the functions
 * in the exported [[Parser]] module.
 *
 * ### Evaluation
 * The language accepted by this parser allows for "sub-expressions" — runs of code that may remain
 * unevaluated, to be executed at a later point.
 *
 * The most direct method of interacting with these, is manually testing whether a particular method
 * has produced a [[Literal]] or a [[Sub]]-expression, and proceeding as appropriate:
 *
 * ```typescript
 * const expr = Parser.expressionOfString('(echo test_cmd) arg'),
 *    cmd = expr.command
 *
 * switch (cmd.type) {
 *    case 'literal': console.log(cmd.value) // ... use stringish val,
 *    case 'subexpression': handleSubexpr(cmd.expr) // or evaluate.
 * }
 * ```
 *
 * However, this can be tedious, if you only wish to evaluate simple expressions. For these
 * situations, I provide a simplified "recursive evaluation" interface: by inverting control, you
 * can allow the parser to become responsible for recursively evaluating [[Sub]] and [[Literal]]
 * values into simple `string`s, as long as you can provide the parser with a simplified
 * handler-callback that will attempt to reduce a single, given sub-expression into a `string`.
 *
 * The entry-point to this auto-evaluation interface is either [[cloneWithEvaluator]], or as a
 * further convenience, the various `eval*` methods below: [[evalCommand]], [[evalPositionals]], and
 * so on.
 *
 * Each of these takes a function — matching the [[evaluator]] signature — that takes an expression,
 * and mechanistically reduces it into a simple string:
 *
 * ```typescript
 * const expr = Parser.expressionOfString('(echo test_cmd) arg')
 *
 * function handleSubexpr(expr) {
 *    if (expr.command === 'echo') {
 *       return expr.positionals.join(' ')
 *    }
 * }
 *
 * console.log("Command:", expr.evalCommand(handleSubexpr))
 * //=> prints "Command: test_cmd"
 * ```
 *
 * (Note, additionally, that within the body of the evaluator above, I've only had to compare
 * `expr.command` to a string, directly. As evaluators receive [[ExpressionEval]]s instead of plain
 * expressions, their bodies need no further shenanigans; any requested expression-properties will
 * recursively evaluate further subexpressions as-needed.)
 *
 * Finally, if you want to access multiple elements of an expression thusly, it may be easier to
 * pre-associate an evaluator with the expression using [[cloneWithEvaluator]], producing an
 * [[ExpressionEval]] directly:
 *
 * ```typescript
 * function handleSubexpr(expr) {
 *    if (expr.command === 'echo') {
 *       return expr.positionals.join(' ')
 *    }
 * }
 *
 * const expr = Parser.expressionOfString('(echo test_cmd) arg'),
 *    ee = expr.cloneWithEvaluator(handleSubexpr)
 *
 * console.log(`Command: ${ee.command}, args: ${ee.positionals.join(' ')}`)
 * //=> prints "Command: test_cmd, args: arg"
 * ```
 *
 * ### Mutation
 * Importantly, these `Expression`s are *mutable* views onto the underlying AST. They're designed to
 * ensure you don't access the data-structure in an unsafe way. In particular, a given `word` in the
 * original input-string can only be *either* a flag-payload, *or* a positional argument — not both.
 *
 * Consider the following:
 *
 * ```typescript
 * const expr = Parser.expressionOfString('hello -f world')
 * ```
 *
 * Does that command have one flag, with the payload `world`? Or an empty (boolean) flag, and a
 * single positional-argument, `world`? My answer, in this interface, is “whichever you observe
 * first”. To be more specific, any method of either [[Expression]] or [[ExpressionEval]] that
 * produces flag-payloads or positional arguments, will *mutate* the data-structure such that
 * subsequent accesses don't see an inconsistent state.
 *
 * Let's look at some examples:
 *
 * ```typescript
 * const expr1 = Parser.expressionOfString('hello -f world')
 *
 * expr1.getFlag('f').value //=> returs 'world';
 * expr1.getPositionals() //=> returns []
 * ```
 *
 * Here, [[getFlag]] mutated `expr1`; and so a subsequent [[getPositionals]] produced a consistent
 * result: there were no remaining positional arguments.
 *
 * ```typescript
 * const expr2 = Parser.expressionOfString('hello -f world')
 *
 * expr2.getPositionals() //=> returns one arg, [{ value: 'world' }]
 * expr2.getFlag('f') //=> returs undefined; because `-f` has no payload
 * ```
 *
 * Here, [[getPositionals]] mutated `expr2`; and so a subsequent [[getFlag]], *again*, produced a
 * consistent result: there was no value available to become the payload of `-f`.
 *
 * This interface is intended to be used in a form wherein the invocations of these methods acts as
 * the specification of the parser's behaviour. For example, one command might expect a flag `-d` or
 * `--dest`, and demand a payload for that flag; whereas another might use `-d` as a boolean, and
 * expect positionals. This way, the behaviour of the parser is flexible enough to handle both of
 * those situations.
 *
 * Finally, note that at least for flags, there's also non-mutative accessors: you can always check
 * whether a flag *exists* using [[hasFlag]].
 */
export declare class Expression extends ExpressionCommon {
    clone(): Expression;
    cloneWithEvaluator(handleSubexpr: evaluator): ExpressionEval;
    readonly command: Literal | Sub<Expression>;
    evalCommand(handleSubexpr: evaluator): string;
    getPositionals(): (Literal | Sub<Expression>)[];
    evalPositionals(handleSubexpr: evaluator): string[];
    forEachFlag(f: (name: string, payload: Literal | Sub<Expression> | undefined, idx: number) => void): void;
    evalEachFlag(handleSubexpr: evaluator, f: (name: string, payload: string | undefined, idx: number) => void): void;
    getFlag(flag: string): Literal | Sub<Expression> | undefined;
    evalFlag(handleSubexpr: evaluator, flag: string): string | undefined;
}
export declare class ExpressionEval extends ExpressionCommon {
    private evaluator;
    constructor(isInternal: sentinel, $expr: $Expressiont, handleSubexpr: evaluator);
    clone(): ExpressionEval;
    readonly command: string;
    getPositionals(): string[];
    forEachFlag(f: (name: string, payload: string | undefined, idx: number) => void): void;
    getFlag(flag: string): string | undefined;
}
export declare class LexBuffer {
    /** @hidden */
    $buf: $buffer;
    constructor(isInternal: sentinel, $buf: $buffer);
    static ofString(str: string): LexBuffer;
    next(): LocatedToken;
    rest(): LocatedToken[];
}
export declare class Position {
    /** @hidden */
    private $pos;
    constructor(isInternal: sentinel, $pos: $position);
    readonly filename: string;
    readonly line: number;
    readonly byte: number;
    readonly character: number;
}
export declare class LocatedToken {
    /** @hidden */
    private $loctok;
    constructor(isInternal: sentinel, $loctok: $token_located);
    readonly $token: $token;
    readonly token: Token;
    readonly id: symbol;
    readonly startLine: number;
    readonly startIdx: number;
    readonly endLine: number;
    readonly endIdx: number;
    readonly body: string | undefined;
}
export declare class Token {
    /** @hidden */
    private $token;
    constructor(isInternal: sentinel, $tok: $token);
    readonly id: symbol;
    readonly body: string | undefined;
    compare(other: Token): boolean;
}
/** @hidden */
interface SemanticMap {
    script: Script;
    expression: Expression;
}
/** @hidden */
declare type SemanticDiscriminator = keyof SemanticMap;
export declare class Checkpoint<D extends SemanticDiscriminator> {
    /** @hidden */
    $cp: $checkpoint;
    private type;
    constructor(isInternal: sentinel, $cp: $checkpoint, type: D);
    /** @hidden */
    producesScript(): this is Checkpoint<'script'>;
    /** @hidden */
    producesExpression(): this is Checkpoint<'expression'>;
    continue<T>(opts?: {
        onAccept?: (val: SemanticMap[D]) => T;
        onFail?: (lastGood: Checkpoint<D>, errorAt: Checkpoint<D>) => T;
    }): T;
    readonly acceptable_token: Token;
    readonly acceptable_tokens: Token[];
    readonly command: Literal | Sub<Expression> | undefined;
    readonly automaton_status: checkpoint_state;
    readonly incoming_symbol_category: 'Terminal' | 'Nonterminal' | undefined;
    readonly incoming_symbol_type: string | undefined;
    readonly incoming_symbol: string | undefined;
    readonly beforeStack: AutomatonStack<D>;
    readonly afterStack: AutomatonStack<D>;
}
export declare class AutomatonStack<D extends SemanticDiscriminator> {
    cp: Checkpoint<D>;
    isAfterStack: boolean;
    constructor(isInternal: sentinel, cp: Checkpoint<D>, isAfterStack: boolean);
    [Symbol.iterator](): IterableIterator<AutomatonElement>;
}
export declare class AutomatonElement {
    /** @hidden */
    private $el;
    constructor(isInternal: sentinel, $el: $element);
    readonly incoming_symbol_category: 'Terminal' | 'Nonterminal';
    readonly incoming_symbol_type: string;
    readonly incoming_symbol: string;
}
declare const _default: {
    OCamlError: typeof OCamlError;
    LexError: typeof LexError;
    Parser: {
        script: typeof script;
        scriptOfString: typeof scriptOfString;
        startScript: typeof startScript;
        startScriptWithString: typeof startScriptWithString;
        expression: typeof expression;
        expressionOfString: typeof expressionOfString;
        startExpression: typeof startExpression;
        startExpressionWithString: typeof startExpressionWithString;
    };
    Script: typeof Script;
    Expression: typeof Expression;
    LexBuffer: typeof LexBuffer;
    LocatedToken: typeof LocatedToken;
    Token: typeof Token;
    Checkpoint: typeof Checkpoint;
};
export default _default;
//# sourceMappingURL=interface.d.ts.map